#!/usr/bin/env python

# $Id: crossCommonLineSearch.py,v 1.13 2005/03/07 23:13:27 xliu Exp $


import sys
import glob
import os
import copy
import time
import random
from math import *
import string
from sys import argv
try:
	from optparse import OptionParser
except:
	from optik import OptionParser
try:
	import mpi
except:
	mpi = None
		
import EMAN


def LOGbegin(ARGV):
	out=open(".emanlog","a")
	b=string.split(ARGV[0],'/')
	ARGV[0]=b[len(b)-1]
	out.write("%d\t%d\t1\t%d\t%s\n" % (os.getpid(),time.time(),os.getppid(),string.join(ARGV," ")))
	out.close()
	return os.getpid()

def LOGInfile(pid,file):
	out=open(".emanlog","a")
	out.write("%d\t%d\t%d\t%s\t\n" % (pid,time.time(),10,file))
	out.close()

def LOGReffile(pid,file):
	out=open(".emanlog","a")
	out.write("%d\t%d\t%d\t%s\t\n" % (pid,time.time(),12,file))
	out.close()

def LOGOutfile(pid,file):
	out=open(".emanlog","a")
	out.write("%d\t%d\t%d\t%s\t\n" % (pid,time.time(),11,file))
	out.close()
	
def LOGend():
	out=open(".emanlog","a")
	out.write("%d\t%d\t2\t-\t\n" % (os.getpid(),time.time()))
	out.close()
	

def parse_command_line():
	usage="Usage: %prog <Raw image filename 1> ... <Raw image filename n> <reference filename> [options]"
	parser = OptionParser(usage=usage)

	#options
	parser.add_option("--rMask",dest="rMask",type="int",help="Mask will be applied to image center. The radial size = rMask, [half image size]",default=0)
	parser.add_option("--rMask1",dest="rMask1",type="int",help="Mask will be applied to solved particle center. The radial size = rMask1, [rMask]",default=0)
	parser.add_option("--startNum",dest="startNumOfRawImages",type="int",help="The starting image number",default=0)
	parser.add_option("--endNum",dest="endNumOfRawImages",type="int",help="The ending image number, (0 is default, all images)",default=0)
	parser.add_option("--numOfRefImages",dest="numOfRefImages",type="int",help="Number of projection images to use for calculating orientation, (5 is default)",default=5)
	parser.add_option("--scatteringFactorFile",dest="sfFileName",type="string",help="File name for scattering factor of raw images",default="")
	parser.add_option("--overSample",dest="FFTOverSampleScale",type="int",help="Scale of over sample FFT, (1 is default, NO over sample)",default=1)
	parser.add_option("--radialStepSize",dest="deltaR",type="float",help="Radial step size for collecting data on common line, (1.0 is default)",default=1.0)
	parser.add_option("--radialStartPoint",dest="RMin",type="int",help="Radial start point for collecting data on common line, (5 is default)",default=5)
	parser.add_option("--radialEndPoint",dest="RMax",type="int",help="Radial end point for collecting data on common line, (35 is default)",default=35)
	#parser.add_option("--maskRadius",dest="maskR",type="int",help="Mask radius to remove background pixels at large radius (default to half of image size)",default=0)
	parser.add_option("--orientationSearchRange",dest="orientationSearchRange",type="float",help="Maximum orientation deviation from a given one, (36.0 is default)",default=36.0)
	parser.add_option("--centerSearchRange",dest="centerSearchRange",type="float",help="Maximum center deviation from a given one, (15.0 is default)",default=15.0)
	parser.add_option("--maxIteration",dest="maxNumOfIteration",type="int",help="Maximum iteration, (1000 is default for coarse search, 150 for refinement)",default=150)
	parser.add_option("--numOfRandomJump",dest="numOfRandomJump",type="int",help="iteration of beginning random jump, (50 is default for coarse search, 20 for refinement)",default=20)
	parser.add_option("--numOfFastShrink",dest="numOfFastShrink",type="int",help="iteration of ending fast shrink, (150 is default for coarse search, 80 for refinement)",default=80)
	parser.add_option("--numOfStartConfigurations",dest="numOfStartConfigurations",type="int",help="starting configurations, (10 is default)",default=10)
	parser.add_option("--searchMode",dest="searchMode",type="int",help="Mode to handle a raw image, (0 is default)",default=0)
	parser.add_option("--scalingMode",dest="scalingMode",type="int",help="Mode to handle a raw image, (0 is default)",default=0)
	parser.add_option("--residualMode",dest="residualMode",type="int",help="Mode to calculate residual, (2 is default)",default=2)
	parser.add_option("--iniParameterFile",dest="rawImageIniParmFN",type="string",help="File name for initial orientation and center of raw images",default="")
	parser.add_option("--refEulerConvention",dest="refEulerConvention",type="string",help="Spcify an euler convention of the refence orientation [eman]",default="eman")
	parser.add_option("--iniCenterOrientationMode",dest="iniCenterOrientationMode",type="string",help="mode to initialize particle center and orientation [random]",default="")
	parser.add_option("--refCenterOrientationMode",dest="refCenterOrientationMode",type="string",help="mode to set reference center and orientation of the particle [copy from ini]",default="")
	parser.add_option("--zScoreCriterion",dest="zScoreCriterion",type="float",help="The criterion of stopping run, (6.0 is default)",default=6.0)
	parser.add_option("--residualCriterion",dest="residualCriterion",type="float",help="The criterion of stopping run, (default: no residual criterion)",default=-1.0)
	parser.add_option("--solutionCenterDiffCriterion",dest="solutionCenterDiffCriterion",type="float",help="The criterion of stopping run, (default: no solution center diff criterion)",default=-1.0)
	parser.add_option("--solutionOrientationDiffCriterion",dest="solutionOrientationDiffCriterion",type="float",help="The criterion of stopping run, (default: no solution orientation diff criterion)",default=-1.0)
	parser.add_option("--maxNumOfRun",dest="maxNumOfRun",type="int",help="Maximum number of re-search, (5 is default)",default=5)
	parser.add_option("--updateHeader", dest="updataHeader",type="int", help="Write the calculated result to the header file, (default is O, NO update)",default=0)
	parser.add_option("--solutionFile",dest="solutionFile",type="string",help="File name for outputing results, (default is Raw image filename + \"-OrientationCenter.dat\")",default="")
	parser.add_option("--mrcSolutionFile",dest="mrcSolutionFile",type="string",help="File name for outputing results in MRC format, (default is Raw image filename + \"-MRC-OrientationCenter.dat\") ",default="")
	parser.add_option("--listFile",dest="listFile",type="string",help="File name for listing good particles in EMAN LST format",default="")
	parser.add_option("--verbose", dest="verbose",type="int", help="verbose level [0-5], (default is O, no screen print out)",default=0)

	(options, args)=parser.parse_args()

	if len(args) < 2:
		parser.print_help()
		sys.exit(-1)
	#elif len(args) == 1:
	#	args = [args[0],'ref.img']
	#elif len(args) > 2:
	#	if (options.verbose > 0) :
	#		print "Warning :: Ignoring the following arguments "
	#	for a in args[2:]: print a
	#	args = [args[0], args[1]]


	for key in parser.option_list[1:]:  #this omits the -h/--help option
		eval_string = "options.%s" % (key.dest)
		value = eval(eval_string)
		if value == None:
			if (options.verbose > 0) :
				print "The option %s must be specified on the command line" % (key)
			sys.exit(-1)

	#remove the input string's spaces, some string may need to convert to lower cases
	options.sfFileName = options.sfFileName.replace(" ", "")
	options.rawImageIniParmFN = options.rawImageIniParmFN.replace(" ", "")
	options.solutionFile = options.solutionFile.replace(" ", "")
	options.mrcSolutionFile = options.mrcSolutionFile.replace(" ", "")
	options.iniCenterOrientationMode = options.iniCenterOrientationMode.lower().replace(" ", "")
	options.refCenterOrientationMode = options.refCenterOrientationMode.lower().replace(" ", "")
	options.refEulerConvention = options.refEulerConvention.lower().replace(" ", "")

	#if options.solutionFile == "":
	#	options.solutionFile = os.path.splitext(args[0])[0] + "-OrientationCenter.dat"

	#if options.mrcSolutionFile == "":
	#	options.mrcSolutionFile = os.path.splitext(args[0])[0] + "-MRC-OrientationCenter.dat"

	#numOfRawImages = EMAN.fileCount(args[0])[0]
	#if options.endNumOfRawImages < 1:
	#	options.endNumOfRawImages = numOfRawImages

	numOfRefImages = EMAN.fileCount(args[-1])[0]
	if options.numOfRefImages > numOfRefImages:
		options.numOfRefImages = numOfRefImages

	if options.startNumOfRawImages < 0:
		options.startNumOfRawImages = 0

	if (options.searchMode) == 0 : # coarse search
		# set default values
		if options.numOfRandomJump == 20 : options.numOfRandomJump = 1000
		if options.numOfFastShrink == 50 : options.numOfFastShrink = 300
		if options.maxNumOfIteration == 150 : options.maxNumOfIteration = 2500
		# check if the given values are less than mins
		if options.numOfRandomJump < 500 : options.numOfRandomJump = 500
		if options.numOfFastShrink <150 : options.numOfFastShrink = 150
		if options.maxNumOfIteration < (options.numOfRandomJump + options.numOfFastShrink + 500) :
			options.maxNumOfIteration = (options.numOfRandomJump + options.numOfFastShrink + 500)
		if options.maxNumOfIteration < 1500 : options.maxNumOfIteration = 1500
	elif (options.searchMode) == 1 : # refinement
		if options.rawImageIniParmFN == "" :
			if (options.verbose > 0) : print "\n\n\tFor refinement, you need to give a file name which specifies orientation and center of each image\n"
			#sys.exit(0)
		# check if the given values are less than mins
		if options.numOfRandomJump < 20 : options.numOfRandomJump = 20
		if options.numOfFastShrink < 50 : options.numOfFastShrink = 50
		if options.maxNumOfIteration < (options.numOfRandomJump + options.numOfFastShrink + 50) :
			options.maxNumOfIteration = (options.numOfRandomJump + options.numOfFastShrink + 50)
		if options.maxNumOfIteration < 120 : options.maxNumOfIteration = 120
	if options.orientationSearchRange > 36 : options.orientationSearchRange = 36
	if options.centerSearchRange > 40 : options.centerSearchRange = 40
	if options.zScoreCriterion < 0 : options.zScoreCriterion = 0
	
	allInputParameters   = "-------------------------------------------------------------------------\n"
	allInputParameters  += "--------------- All input parameters are listed as follows --------------\n"
	allInputParameters  += "-------------------------------------------------------------------------\n"
	allInputParameters  += "%40s :: %s \n" % ("Raw image file name",args[0:-1])
	allInputParameters  += "%40s :: %s \n" % ("Reference image file name",args[-1])
	if options.sfFileName != "" : allInputParameters  += "%40s :: %s \n" % ("Scattering factor file name",options.sfFileName)
	if options.searchMode == 1  : allInputParameters  += "%40s :: %s \n" % ("Raw image initial parameter file name",  options.rawImageIniParmFN)
	allInputParameters  += "%40s :: %s \n" % ("Starting number of raw images", options.startNumOfRawImages)
	allInputParameters  += "%40s :: %s \n" % ("Ending number of raw images", options.endNumOfRawImages)
	allInputParameters  += "%40s :: %s \n" % ("Number of reference images", options.numOfRefImages)
	allInputParameters  += "%40s :: %d \n" % ("Mask radius (pixels)", options.rMask)
	allInputParameters  += "%40s :: %s \n" % ("Scale of over sample FFT", options.FFTOverSampleScale)
	allInputParameters  += "%40s :: %s \n" % ("Maximum number of iteration", options.maxNumOfIteration)
	allInputParameters  += "%40s :: %s \n" % ("Iteration number of random jump", options.numOfRandomJump)
	allInputParameters  += "%40s :: %s \n" % ("Iteration number of fast shrink", options.numOfFastShrink)
	allInputParameters  += "%40s :: %s \n" % ("Center search range", options.centerSearchRange)
	allInputParameters  += "%40s :: %s \n" % ("Orientation search range", options.orientationSearchRange)
	allInputParameters  += "%40s :: %s \n" % ("Radial Step size of a common line", options.deltaR)
	allInputParameters  += "%40s :: %s \n" % ("Radial start point of a common line", options.RMin)
	allInputParameters  += "%40s :: %s \n" % ("Radial end point of a common line", options.RMax)
	allInputParameters  += "%40s :: %s \n" % ("Mode to handle a raw image", options.searchMode)
	allInputParameters  += "%40s :: %s \n" % ("Mode to calculate residual", options.residualMode)

	allInputParameters  += "%40s :: %s \n" % ("Maximum number of re-run", options.maxNumOfRun)
	allInputParameters  += "%40s :: %s%s \n" % ("The criteria of discriminating data", "z score = ", options.zScoreCriterion)
	if options.residualCriterion > 0 : allInputParameters  += "%40s :: %s%s \n" % ("", "residual threshold = ", options.residualCriterion)
	else : allInputParameters  += "%40s :: %s \n" % ("", "residual threshold = N/A")
	if options.solutionCenterDiffCriterion > 0 : allInputParameters  += "%40s :: %s%s \n" % ("", "center difference = ", options.solutionCenterDiffCriterion)
	else : allInputParameters  += "%40s :: %s \n" % ("", "center difference = N/A")
	if options.solutionOrientationDiffCriterion > 0 : allInputParameters  += "%40s :: %s%s \n" % ("", "orientation difference = ", options.solutionOrientationDiffCriterion)
	else : allInputParameters  += "%40s :: %s \n" % ("", "orientation difference = N/A")

	printOnScreen   = allInputParameters
	printOnScreen  += "-------------------------------------------------------------------------\n"
	printOnScreen  += "%40s :: %s \n" % ("Solution file name", options.solutionFile)
	printOnScreen  += "-------------------------------------------------------------------------\n"

	writeToSolutionFile  = allInputParameters
	writeToSolutionFile += "-------------------------------------------------------------------------\n\n"
	writeToSolutionFile += "%6s%8s%7s%7s%7s%7s%10s%34s%16s%24s%9s%12s\n" % ("######","X","Y","Alt","Az","Phi","residual","  sigma      Z  Z-crti Run# Status","ref center", "ref orientation", "XY diff","angle diff")
	writeToSolutionFile += "------------------------------------------------------------------------------------------------------------------------\n"
	if options.verbose > 0 and (not mpi or (mpi and mpi.rank==0)) : #print out the input parameters on screen
		print "%s " % (printOnScreen)

	if options.solutionFile and options.startNumOfRawImages == 0 : #write the input parameters to the solution file
		solutionFile = open(options.solutionFile, "w")
		solutionFile.write("%s" % (writeToSolutionFile))
		solutionFile.close()


	return (options,args)


def CTFPeakZeroPosition(defocus,ampcontrast,voltage,Cs,order=1):
	"""Calculate the CTF peak, zero positions using analytic method. defocus should be negative for underfocus"""
	wl=12.2639/sqrt(voltage*1000.0+.97845*voltage*voltage)
	#gamma=2*pi*(2.5e6*Cs*wl*wl*wl*s*s*s*s+5000.*defocus*wl*s*s)
	Qphi = -asin(ampcontrast)
	
	# ctf = sin(gamma+Qphi) = sin(a*s^4+b*s^2+c)
	a = 2*pi*2.5e6*Cs*wl*wl*wl
	b = 2*pi*5000.*defocus*wl
	c = Qphi
	#
	# ctf=0 when a*s^4+b*s^2+c = -n*pi
	#	i.e. s = sqrt(((-b-sqrt(b*b-4*a*(c+n*pi)))/(2*a)))
	# ctf=peak when a*s^4+b*s^2+c = -n*pi-pi/2
	#	i.e. s = sqrt(((-b-sqrt(b*b-4*a*(c+n*pi+pi/2)))/(2*a)))
	#
	try:
		szero = sqrt(((-b-sqrt(b*b-4*a*(c+order*pi)))/(2*a)))
		speak = sqrt(((-b-sqrt(b*b-4*a*(c+(order-1)*pi+pi/2)))/(2*a)))
	except ValueError:
		szero = 0
		speak = 0
	return (szero, speak)


def main():
	EMAN.appinit(sys.argv)
	if sys.argv[-1].startswith("usefs="): sys.argv = sys.argv[:-1]	# remove the runpar fileserver info

	(options,args) =  parse_command_line()

	inputParm = EMAN.ccmlInputParm()
	sf = EMAN.XYData()
	if options.sfFileName != "" :
		readsf = sf.readFile(options.sfFileName)
		if ((readsf == -1) and (options.verbose > 0)) :
			print "The file of scattering factor does NOT exist"
	inputParm.scateringFactor = sf

	startNumOfRawImages = options.startNumOfRawImages
	#endNumOfRawImages = options.endNumOfRawImages

	refImageFileName = args[-1]
	numOfRefImages = options.numOfRefImages
	solutionFile = options.solutionFile

	# write log info to .emanlog file so that eman program can browse the history
	if not mpi or (mpi and mpi.rank==0): 
		pid = LOGbegin(sys.argv)
		for f in args[0:-1]: LOGInfile(pid,f)
		LOGReffile(pid,args[-1])
		if options.solutionFile: LOGOutfile(pid,options.solutionFile)
		if options.listFile: LOGOutfile(pid,options.listFile)
		if options.mrcSolutionFile: LOGOutfile(pid,options.mrcSolutionFile)

	inputParm.FFTOverSampleScale = options.FFTOverSampleScale
	inputParm.deltaR = options.deltaR
	inputParm.RMin = options.RMin
	inputParm.RMax = options.RMax
	inputParm.searchMode = options.searchMode
	inputParm.scalingMode = options.scalingMode
	inputParm.residualMode = options.residualMode
	inputParm.rawImageIniParmFN = options.rawImageIniParmFN

	inputParm.maxNumOfRun = options.maxNumOfRun
	inputParm.zScoreCriterion = options.zScoreCriterion
	inputParm.residualCriterion = options.residualCriterion
	inputParm.solutionCenterDiffCriterion = options.solutionCenterDiffCriterion
	inputParm.solutionOrientationDiffCriterion = options.solutionOrientationDiffCriterion/180.0*pi
	inputParm.maxNumOfIteration = options.maxNumOfIteration
	inputParm.numOfRandomJump = options.numOfRandomJump
	inputParm.numOfFastShrink = options.numOfFastShrink
	inputParm.numOfStartConfigurations = options.numOfStartConfigurations
	inputParm.orientationSearchRange = options.orientationSearchRange/180.0*pi
	inputParm.centerSearchRange = options.centerSearchRange

#	inputParm.refImages = EMAN.readImages(refImageFileName,0,numOfRefImages-1)  #this kind of assignment does NOT work, use the follows
	tempVectorOfRefImages = EMAN.readImages(refImageFileName,0,numOfRefImages-1)
	inputParm.refImages = tempVectorOfRefImages  #you have to go this way, otherwise it can NOT pass parameter correctly
	refImageSizeX = tempVectorOfRefImages[0].xSize()
	refImageSizeY = tempVectorOfRefImages[0].ySize()
	inputParm.numOfRefImages = options.numOfRefImages
	inputParm.refEulerConvention = options.refEulerConvention
	#maskR = options.maskR
	#if (maskR<=0): maskR = refImageSizeY/2
	
	inputParm.verbose = options.verbose
	verbose = options.verbose
	#verboseSolution = options.verboseSolution

	updataHeader = options.updataHeader
	solutionFile = options.solutionFile
	mrcSolutionFile = options.mrcSolutionFile
	iniCenterOrientationMode = options.iniCenterOrientationMode
	refCenterOrientationMode = options.refCenterOrientationMode

	createRefCenterOrientation = 0
	if options.solutionCenterDiffCriterion > 0 or options.solutionOrientationDiffCriterion > 0 : #if this criterion is specified, the ref orientaiton and center need to be given
		if not (iniCenterOrientationMode == "iniparmfile" or iniCenterOrientationMode == "headerfile" or \
			refCenterOrientationMode == "iniparmfile" or refCenterOrientationMode == "headerfile") :
			createRefCenterOrientation = 1

	rawImages = []
	if not mpi or (mpi and mpi.rank==0):
		for imgfile in args[0:-1]:
			imgnum = EMAN.fileCount(imgfile)[0]
			for i in range(imgnum): rawImages.append((imgfile, i))
	if mpi: rawImages = mpi.bcast(rawImages)
	
	endNumOfRawImages = options.endNumOfRawImages
	if endNumOfRawImages <=0  or endNumOfRawImages > len(rawImages):
		endNumOfRawImages = len(rawImages)

	numRawImages = endNumOfRawImages - startNumOfRawImages

	if mpi:
		ptclset = range(startNumOfRawImages + mpi.rank, endNumOfRawImages, mpi.size)
	else:
		ptclset = range(startNumOfRawImages, endNumOfRawImages)
	
	solutions = []

	rMask = options.rMask        #mask size is given
	if options.rMask <= 0 : rMask = refImageSizeY/2   #mask size = half image size
	
	rMask1 = options.rMask1             #output tnf mask size is given
	if options.rMask1 <= 0 : rMask1 = rMask    #output tnf mask size = half image size

	
	rawImage = EMAN.EMData()
	inputParm.rawImage = rawImage
	#for rawImgSN in ptclset:
	for index in range(len(ptclset)):
		rawImgSN = ptclset[index]
		inputParm.thisRawImageSN = rawImgSN
		if mpi: print "rank %d: %d in %d-%d (%d in %d-%d)" % (mpi.rank, rawImgSN, startNumOfRawImages, endNumOfRawImages, index, 0, len(ptclset))
		rawImage.readImage(rawImages[rawImgSN][0], rawImages[rawImgSN][1])

		if rawImage.xSize() != refImageSizeX:
			print "Image %s size (%dx%d) is different to reference image %s size (%dx%d), I will quit" % \
				(rawImages[rawImgSN][0],rawImage.xSize(),rawImage.ySize(),refImageFileName,refImageSizeX,refImageSizeY)
			sys.exit()

		rawImage.applyMask(rMask, 6) #apply mask type 6 [edge mean value] to raw image, center will be image center
		rawImage.getEuler().setSym("icos")
		if rawImage.hasCTF() == 1:
			ctfParm = rawImage.getCTF()
			#inputParm.zScoreCriterion = options.zScoreCriterion + atan(abs(ctfParm[0])-1.5)/(pi/4) +0.59 #adjust zScore criterion -0.6 --> +1.2, 1.5, 2.0
			#inputParm.numOfRefImages = int(min(numOfRefImages, max(numOfRefImages*exp(-(abs(ctfParm[0])/2.0-0.15))+0.5, 5.0))) # adjust maxNumOfRun, the min is 2


		#initialize Center And Orientation by ont of the following modes

		if iniCenterOrientationMode == "iniparmfile" : inputParm.initializeCenterAndOrientationFromIniParmFile() # need to set "refEulerConvention"
		elif iniCenterOrientationMode == "headerfile" : inputParm.initializeCenterAndOrientationFromParticle() # need to set "refEulerConvention"
		else : inputParm.initializeCenterAndOrientationFromRandom()  # default is random orientation and physical center

		#set the refence Center And Orientation by ont of the following modes

		if refCenterOrientationMode == "iniparmfile" : inputParm.setRefCenterAndOrientationFromIniParmFile() # need to set "refEulerConvention"
		elif refCenterOrientationMode == "headerfile" : inputParm.setRefCenterAndOrientationFromParticle() # need to set "refEulerConvention"
		else : inputParm.setRefCenterAndOrientationFromInitializedParms() # default is copy the initial center and orientation

		inputParm.thisRawImageSN = rawImgSN

		if createRefCenterOrientation or (inputParm.searchMode == 1): #if need reference, do this extra search first
			inputParm.zScoreCriterion = options.zScoreCriterion
			inputParm.residualCriterion = options.residualCriterion
			inputParm.solutionCenterDiffCriterion = -1.0
			inputParm.solutionOrientationDiffCriterion = -1.0
			############################################################################
			############ execute cross common line search for reference ################
			############################################################################
			outputParm  = rawImage.crossCommonLineSearch(inputParm)
			############################################################################
			############################################################################

			inputParm.zScoreCriterion = 2.0
			inputParm.residualCriterion = options.residualCriterion
			inputParm.solutionCenterDiffCriterion = options.solutionCenterDiffCriterion
			inputParm.solutionOrientationDiffCriterion = options.solutionOrientationDiffCriterion/180.0*pi
			if iniCenterOrientationMode == "iniparmfile" : inputParm.initializeCenterAndOrientationFromIniParmFile() # need to set "refEulerConvention"
			elif iniCenterOrientationMode == "headerfile" : inputParm.initializeCenterAndOrientationFromParticle() # need to set "refEulerConvention"
			else : inputParm.initializeCenterAndOrientationFromRandom()  # default is random orientation and physical center
			#inputParm.initializeCenterAndOrientationFromRandom()  # default is random orientation and physical center
			inputParm.setRefCenterAndOrientationFromInput(outputParm.cx, outputParm.cy, outputParm.q)

		############################################################################
		############ execute cross common line search for final result #############
		############################################################################
		outputParm  = rawImage.crossCommonLineSearch(inputParm)
		############################################################################
		############################################################################
		if outputParm.passAllCriteria: 	passfail = "pass"
		else: passfail = "fail"

		print "Final result:", outputParm.cx, outputParm.cy, outputParm.alt*180/pi, outputParm.az*180/pi, outputParm.phi*180/pi, passfail

		if options.solutionFile:
			#outputParm.convertCalculatedOrientationToAsymmetricUnit()
			#outputParm.convertIniOrientationToAsymmetricUnit()
			#outputParm.convertRefOrientationToAsymmetricUnit()

			#outputParm.calculateDifferenceWithIniParm()
			if iniCenterOrientationMode == "iniparmfile" or iniCenterOrientationMode == "headerfile" or \
			   refCenterOrientationMode == "iniparmfile" or refCenterOrientationMode == "headerfile" or \
			   createRefCenterOrientation : 
				outputParm.calculateDifferenceWithRefParm()
			outputParm.outputResult(solutionFile)

		solutions.append( (rawImages[rawImgSN][0], rawImages[rawImgSN][1], outputParm.alt, outputParm.az, outputParm.phi, \
						   outputParm.cx, outputParm.cy, outputParm.passAllCriteria) )
		
		sys.stdout.flush()

	if mpi:
		if options.verbose: 
			print "rank %d: done and ready to output" % (mpi.rank)
			sys.stdout.flush()
		mpi.barrier()
		#print "rank %d: %s" % (mpi.rank, solutions)
		if mpi.rank==0:
			for r in range(1,mpi.size):
				msg, status = mpi.recv(source = r, tag = r)
				solutions += msg
			def ptcl_cmp(x, y):
				eq = cmp(x[0], y[0])
				if not eq: return cmp(x[1],y[1])
				else: return eq
			solutions.sort(ptcl_cmp)
		else:
			mpi.send(solutions, 0, tag = mpi.rank)

	if not mpi or (mpi and mpi.rank==0):
		if options.listFile:
			lFile = open(options.listFile, "w")
			lFile.write("#LST\n")
			for i in solutions:
				if i[-1]: 
					lFile.write("%d\t%s\t%g\t%g\t%g\t%g\t%g\n" % (i[1], i[0], i[2]*180.0/pi, i[3]*180.0/pi, i[4]*180.0/pi, i[5], i[6]))
			lFile.close()
		if options.mrcSolutionFile:
			outFile = open(options.mrcSolutionFile, "w")
			for i in solutions:
				if i[-1]:
					#rawImage.readImage(i[0], i[1], 1)
					rawImage.readImage(i[0], i[1])
					thisEu = EMAN.Euler(i[2], i[3], i[4])
					thisEu.convertToMRCAngle()
					alt = thisEu.alt_MRC()*180.0/pi
					az  = thisEu.az_MRC()*180.0/pi
					phi = thisEu.phi_MRC()*180.0/pi
		
					cx  = i[5]
					cy  = i[6]
					dx = cx - rawImage.xSize()/2
					dy = cy - rawImage.ySize()/2
					rawImage.applyMask(rMask1,6,dx,dy,0) #apply mask type 4 [outside=0] to raw image, center will be the solved center
					#tnfFileName = "%s-%d.tnf" % (os.path.basename(os.path.splitext(rawImages[rawImgSN][0])[0]), rawImages[rawImgSN][1])
					prefix = os.path.dirname(options.mrcSolutionFile).replace(" ", "")
					if prefix != "" : prefix = prefix + "/"
					tnfFileName = "%s%s-%d.tnf" % (prefix,os.path.basename(os.path.splitext(i[0])[0]), i[1])
					rawFFT = rawImage.doFFT()
					rawFFT.writeImage(tnfFileName,0)  #tnf file no header information, it is a pure FFT of raw image file
		
					outFile.write("%s\n" % (os.path.abspath(tnfFileName)))
					outFile.write(" %d, %.4f, %.4f, %.4f, %.4f, %.4f, 0.0\n" % (0, alt, az, phi, cy, cx))
			outFile.close()
		if updataHeader:
			for i in solutions:
				rawImage.readImage(i[0], i[1], 1)
				if options.verbose:
					cx  = rawImage.get_center_x()
					cy  = rawImage.get_center_y()
					alt = rawImage.alt()
					az  = rawImage.az()
					phi = rawImage.phi()
					print "Update header: %s %d\t%7.5f  %7.5f  %7.2f  %7.2f  %7.2f => %7.5f  %7.5f  %7.2f  %7.2f  %7.2f" % \
						(i[0], i[1], alt*180.0/pi, az*180.0/pi, phi*180.0/pi, cx, cy, i[2]*180.0/pi, i[3]*180.0/pi, i[4]*180.0/pi, i[5], i[6])
				rawImage.setRAlign(i[2], i[3], i[4])
				rawImage.set_center_x(i[5])
				rawImage.set_center_y(i[6])
				imgtype = EMAN.EMData.ANY
				rawImage.writeImage(i[0], i[1], imgtype, 1)
	if not mpi or (mpi and mpi.rank==0): LOGend()

if __name__== "__main__":
	main()



